@@ -17,12 +17,16 @@
  */
 #define PAM_MUST_FAIL_CODE        PAM_PERM_DENIED
 
+/* impression codes - this gives some sense to the logical choices */
+
+#define _PAM_UNDEF     0
+#define _PAM_POSITIVE +1
+#define _PAM_NEGATIVE -1
+
 static int _pam_dispatch_aux(pam_handle_t *pamh, int flags, struct handler *h)
 {
-    int ret, returned;
-    int impression = PAM_PERM_DENIED;    /* carries the general impression */
-    int result = PAM_PERM_DENIED;        /* this is what is to be returned */
-    int got_one = 0;
+    int impression = _PAM_UNDEF;
+    int status = PAM_MUST_FAIL_CODE;
 
     IF_NO_PAMH("_pam_dispatch_aux", pamh, PAM_SYSTEM_ERR);
 
@@ -32,124 +36,111 @@ static int _pam_dispatch_aux(pam_handle_t *pamh, int flags, struct handler *h)
 	(void) pam_get_item(pamh, PAM_SERVICE, (const void **)&service);
 	_pam_log_error("no modules loaded for `%s' service"
 		       , service ? service:"<unknown>" );
-
-	D(("no loaded modules for `%s' service"
-	   , service ? service:"<unknown>"));
 	service = NULL;
-	return PAM_PERM_DENIED;
+	return PAM_MUST_FAIL_CODE;
     }
 
-    /* loop through all modules */
-    while (h) {
+    /* Loop through module logic stack */
+    for ( ; h != NULL ; h = h->next ) {
+	int retval, action;
 
-#ifdef DEBUG
-	{
-	    int i;
-	    D(("argc=%d\nargv=",h->argc));
-	    for (i=0; i<h->argc; ++i) {
-		D(("\t%s",h->argv[i]));
-	    }
-	}
-#endif
-	D(("user is %s", pamh->user));
+	/* attempt to call the module */
 	if (h->func == NULL) {
-	    D(("module function is not defined, indicating fail"));
-	    returned = PAM_MUST_FAIL_CODE;   /* mapped to other error above */
+	    D(("module function is not defined, indicating failure"));
+	    retval = PAM_MODULE_UNKNOWN;
 	} else {
 	    D(("passing control to module..."));
-	    returned = h->func(pamh, flags, h->argc, h->argv);
+	    retval = h->func(pamh, flags, h->argc, h->argv);
 	    D(("module returned: %s", pam_strerror(pamh, returned)));
 	    if (h->must_fail) {
 		D(("module poorly listed in pam.conf; forcing failure"));
-		returned = PAM_MUST_FAIL_CODE;  /* mapped to other error ^^ */
+		retval = PAM_MUST_FAIL_CODE;
 	    }
 	}
 
-	/*
-	 * some returns can be treated like others: ret is less specific
-	 * than returned, and indicates success as PAM_SUCCESS.
-	 */
+	/* verify that the return value is a valid one */
+	if (retval < PAM_SUCCESS || retval >= _PAM_RETURN_VALUES) {
+	    retval = PAM_MUST_FAIL_CODE;
+	    action = _PAM_ACTION_BAD;
+	} else {
+	    action = h->actions[retval];
+	}
 
-	switch (returned) {
-	case PAM_NEW_AUTHTOK_REQD:         /* equivalent to success */
-	    ret = PAM_SUCCESS;
-	    D(("treating `%s' as PAM_SUCCESS", pam_strerror(pamh, returned)));
+	/* decide what to do */
+	switch (action) {
+	case _PAM_ACTION_RESET:
+	    impression = _PAM_UNDEF;
+	    status = PAM_MUST_FAIL_CODE;
 	    break;
-	default:
-	    ret = returned;              /* no mapping so take raw value */
-	}
 
-	if (ret != PAM_IGNORE) {
-	    /* we should not ignore this result */
-
-	    if (ret == PAM_SUCCESS) {
-
-		if (!got_one) {              /* this is our first success */
-		    impression = ret;        /* keep impression of success */
-		    result = returned;       /* not PERM_DENIED now */
-		    got_one = 1;
-		}
-
-		if (impression == PAM_SUCCESS      /* not failed a required */
-		    && h->control_flag == PAM_F_SUFFICIENT /* is sufficient */
-		    && ret == PAM_SUCCESS ) {      /* and it was successful */
-		    D(("sufficient for success"));
-		    break;      /* success! */
-		}
-
-	    } else {
-		/* this is a failure... */
+	case _PAM_ACTION_OK:
+	case _PAM_ACTION_DONE:
+	    if ( impression == _PAM_UNDEF
+		 || (impression == _PAM_POSITIVE && status == PAM_SUCCESS) ) {
+		impression = _PAM_POSITIVE;
+		status = retval;
+	    }
+	    if ( impression == _PAM_POSITIVE && action == _PAM_ACTION_DONE ) {
+		goto decision_made;
+	    }
+	    break;
 
+	case _PAM_ACTION_BAD:
+	case _PAM_ACTION_DIE:
 #ifdef PAM_FAIL_NOW_ON
-		if (ret == PAM_ABORT) {
-		    /*
-		     * This is the result of an exceptional condition.  The
-		     * module wants to stop talking to the user immediately.
-		     */
-		    got_one = 1;
-		    impression = result = PAM_PERM_DENIED;
-		    D(("pam failing now"));
-		    break;
-		}
+	    if ( retval == PAM_ABORT ) {
+		impression = _PAM_NEGATIVE;
+		status = PAM_PERM_DENIED;
+		goto decision_made;
+	    }
 #endif /* PAM_FAIL_NOW_ON */
+	    if ( impression != _PAM_NEGATIVE ) {
+		impression = _PAM_NEGATIVE;
+		status = retval;
+	    }
+	    if ( action == _PAM_ACTION_DIE ) {
+		goto decision_made;
+	    }
+	    break;
 
-		if ( h->control_flag == PAM_F_REQUIRED
-		     || h->control_flag == PAM_F_CRITICAL ) {
-		    D(("significant failure"));
-		    if (!got_one || impression == PAM_SUCCESS) {
-			D(("the recorded failure"));
-			impression = ret;    /* set to fail */
-			result = returned;   /* this takes the value of
-						the first fail on a
-						PAM_REQUIRED module */
-		    }
-		    got_one = 1;        /* only care about significant fails */
-		    if ( h->control_flag == PAM_F_CRITICAL ) {
-			D(("critical module has failed..."));
-			break;
-		    }
-		} else {
-		    D(("ignoring failure"));
-		}
+	case _PAM_ACTION_IGNORE:
+	    break;
+
+        /* if we get here, we expect action is a positive number --
+           this is what the ...JUMP macro checks. */
+
+	default:
+	    if ( _PAM_ACTION_IS_JUMP(action) ) {
+		/* this means that we need to skip #action stacked modules */
+		do {
+ 		    h = h->next;
+		} while ( --action > 0 && h != NULL );
+
+		/* note if we try to skip too many modules action is
+                   still non-zero and we snag the next if. */
+	    }
+
+	    /* this case is a syntax error: we can't succeed */
+	    if (action) {
+		D(("action syntax error"));
+		impression = _PAM_NEGATIVE;
+		status = PAM_MUST_FAIL_CODE;
 	    }
 	}
-
-	h = h->next;
     }
 
-    if (got_one) {
-	D(("returning %s", pam_strerror(pamh, result)));
-	return result;
+decision_made:     /* by getting  here we have made a decision */
+
+    /* Sanity check */
+    if ( status == PAM_SUCCESS && impression != _PAM_POSITIVE ) {
+	D(("caught on sanity check -- this is probably a config error!"));
+	status = PAM_MUST_FAIL_CODE;
     }
 
-/* this is if there are only PAM_OPTIONAL entries (or none at all)
-   and they all fail! Is there a better error to return than this?  */
-
-    D(("falling through to default failure"));
-    return PAM_PERM_DENIED;
+    /* We have made a decision about the modules executed */
+    return status;
 }
 
-
 int _pam_dispatch(pam_handle_t *pamh, int flags, int choice)
 {
     struct handler *h = NULL;
