@@ -1,7 +1,7 @@
 /* pam_item.c */
 
 /*
- * $Id: pam_item.c,v 1.8 1997/02/15 15:58:49 morgan Exp $
+ * $Id: pam_item.c,v 1.8 1997/02/15 15:58:49 morgan Exp morgan $
  *
  * $Log: pam_item.c,v $
  * Revision 1.8  1997/02/15 15:58:49  morgan
@@ -274,8 +274,7 @@ int pam_get_user(pam_handle_t *pamh, const char **user, const char *prompt)
          * correctly free()'d here
 	 */
 
-	_pam_drop(resp->resp);
-	_pam_drop(resp);
+	_pam_drop_reply(resp, 1);
     }
 
     return retval;        /* pass on any error from conversation */
