@@ -1,7 +1,7 @@
 /* pam_mail module */
 
 /*
- * $Id: pam_env.c,v 1.1 1997/04/05 06:42:35 morgan Exp $
+ * $Id: pam_env.c,v 1.1 1997/04/05 06:42:35 morgan Exp morgan $
  *
  * Written by Dave Kinchlea <kinch@kinch.ark.com> 1997/01/31
  * Inspired by Andrew Morgan <morgan@parc.power.net, who also supplied the 
@@ -50,7 +50,11 @@
 #define DEFAULT_CONF_FILE       "/etc/security/pam_env.conf"
 #endif
 
+#define _GNU_SOURCE
+#include <features.h>
+
 #include <ctype.h>
+#include <errno.h>
 #include <pwd.h>
 #include <stdarg.h>
 #include <stdio.h>
@@ -113,19 +117,6 @@ static char quote='Z';
 
 /* some syslogging */
 
-static char *xstrdup(const char *x)
-{
-    if (x) {
-	char *y;
-
-	y = (char *) malloc(strlen(x)+1);
-	if (y)
-	    strcpy(y,x);
-	return y;
-    } else
-	return NULL;
-}
-
 static void _log_err(int err, const char *format, ...)
 {
     va_list args;
@@ -156,7 +147,7 @@ static int _pam_parse(int flags, int argc, const char **argv, char **conffile)
 	if (!strcmp(*argv,"debug"))
 	    ctrl |= PAM_DEBUG_ARG;
 	else if (!strncmp(*argv,"conffile=",9)) {
-	    *conffile = xstrdup(9+*argv);
+	    *conffile = x_strdup(9+*argv);
 	    if (*conffile != NULL) {
 		D(("new Configuration File: %s", *conffile));
 		ctrl |= PAM_NEW_CONF_FILE;
@@ -493,8 +484,10 @@ static int _expand_arg(pam_handle_t *pamh, char **value)
 		      
   char type, tmpval[BUF_SIZE]; /* No unexpanded variable can be bigger than BUF_SIZE */
   char tmp[MAX_ENV];   /* I know this shouldn't be hard-coded but it's so 
-			* much easier this way 
-			*/
+			* much easier this way */
+
+  D(("Remember to initialize tmp!"));
+  tmp[0] = '\0';
 
   /* 
    * (possibly non-existent) environment variables can be used as values
