@@ -1,7 +1,7 @@
 /* pam_lastlog module */
 
 /*
- * $Id: pam_lastlog.c,v 1.2 1997/02/15 17:18:21 morgan Exp $
+ * $Id: pam_lastlog.c,v 1.3 1997/04/05 06:18:21 morgan Exp morgan $
  *
  * Written by Andrew Morgan <morgan@parc.power.net> 1996/3/11
  *
@@ -10,6 +10,9 @@
  * present (login) service.
  *
  * $Log: pam_lastlog.c,v $
+ * Revision 1.3  1997/04/05 06:18:21  morgan
+ * removed xstrdup - unused
+ *
  * Revision 1.2  1997/02/15 17:18:21  morgan
  * removed fixed buffer in logging
  *
@@ -19,7 +22,12 @@
  */
 
 #include <fcntl.h>
-#include <lastlog.h>
+#include <time.h>
+#ifdef HAVE_UTMP_H
+# include <utmp.h>
+#else
+# include <lastlog.h>
+#endif
 #include <pwd.h>
 #include <stdlib.h>
 #include <stdarg.h>
@@ -33,6 +41,23 @@
 #include <pwdb/pwdb_public.h>                /* use POSIX front end */
 #endif
 
+#ifdef __hpux
+# ifndef _PATH_LASTLOG
+#  define _PATH_LASTLOG "/usr/adm/lastlog"
+# endif /* _PATH_LASTLOG */
+# ifndef UT_HOSTSIZE
+#  define UT_HOSTSIZE 16
+# endif /* UT_HOSTSIZE */
+# ifndef UT_LINESIZE
+#  define UT_LINESIZE 12
+# endif /* UT_LINESIZE */
+struct lastlog {
+    time_t  ll_time;
+    char    ll_line[UT_LINESIZE];
+    char    ll_host[UT_HOSTSIZE];            /* same as in utmp */
+};
+#endif /* __hpux */
+
 /* XXX - time before ignoring lock. Is 1 sec enough? */
 #define LASTLOG_IGNORE_LOCK_TIME     1
 
@@ -64,19 +89,6 @@
 
 /* some syslogging */
 
-static char *xstrdup(const char *x)
-{
-    char *y;
-
-    if (x) {
-	y = (char *) malloc(strlen(x)+1);
-	if (y)
-	    strcpy(y,x);
-	return y;
-    } else
-	return NULL;
-}
-
 static void _log_err(int err, const char *format, ...)
 {
     va_list args;
@@ -99,7 +111,7 @@ static void _log_err(int err, const char *format, ...)
 
 static int _pam_parse(int flags, int argc, const char **argv)
 {
-    int ctrl=(LASTLOG_DATE|LASTLOG_HOST|LASTLOG_LINE|LASTLOG_NEVER);
+    int ctrl=(LASTLOG_DATE|LASTLOG_HOST|LASTLOG_LINE);
 
     /* does the appliction require quiet? */
     if (flags & PAM_SILENT) {
@@ -121,6 +133,8 @@ static int _pam_parse(int flags, int argc, const char **argv)
 	    ctrl |= ~LASTLOG_HOST;
 	} else if (!strcmp(*argv,"silent")) {
 	    ctrl |= LASTLOG_QUIET;
+	} else if (!strcmp(*argv,"never")) {
+	    ctrl |= LASTLOG_NEVER;
 	} else {
 	    _log_err(LOG_ERR,"unknown option; %s",*argv);
 	}
@@ -151,12 +165,12 @@ static int converse(pam_handle_t *pamh, int ctrl, int nargs
 
 	if (retval != PAM_SUCCESS && (ctrl & LASTLOG_DEBUG) ) {
 	    _log_err(LOG_DEBUG, "conversation failure [%s]"
-		     , pam_strerror(retval));
+		     , pam_strerror(pamh, retval));
 	}
 
     } else {
 	_log_err(LOG_ERR, "couldn't obtain coversation function [%s]"
-		 , pam_strerror(retval));
+		 , pam_strerror(pamh, retval));
     }
 
     D(("ready to return from module conversation"));
@@ -179,12 +193,15 @@ static int make_remark(pam_handle_t *pamh, int ctrl, const char *remark)
 	retval = converse(pamh, ctrl, 1, mesg, &resp);
 
 	msg[0].msg = NULL;
+	if (resp) {
+	    _pam_drop_reply(resp, 1);
+	}
     } else {
 	D(("keeping quiet"));
 	retval = PAM_SUCCESS;
     }
 
-    D(("returning %s", pam_strerror(retval)));
+    D(("returning %s", pam_strerror(pamh, retval)));
     return retval;
 }
 
@@ -232,12 +249,11 @@ static int last_login_date(pam_handle_t *pamh, int announce, uid_t uid)
 	(void) fcntl(last_fd, F_SETLK, &last_lock);        /* unlock */
 
 	if (!win) {
-	    D(("bad news reading the %s file for uid=%d", _PATH_LASTLOG, uid));
+	    D(("First login for user uid=%d", _PATH_LASTLOG, uid));
 	    if (announce & LASTLOG_DEBUG) {
-		_log_err(LOG_DEBUG, "unable to read %s for uid %d"
-			 , _PATH_LASTLOG, uid);
+		_log_err(LOG_DEBUG, "creating lastlog for uid %d", uid);
 	    }
-	    return PAM_SESSION_ERR;
+	    memset(&last_login, 0, sizeof(last_login));
 	}
 
 	/* rewind */
