@@ -77,7 +77,7 @@ char *crypt(const char *key, const char *salt);
 #include <string.h>
 #include <stdarg.h>
 #include <malloc.h>
-#include <security/pam_misc.h>
+#include <security/_pam_macros.h>
 
 #ifndef LINUX    /* AGM added this as of 0.2 */
 #include <security/pam_appl.h>
@@ -304,7 +304,7 @@ int pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv)
     pflags = _do_getpass(usrname,&miscptr);
     if(pflags & PPW_NOSUCHUSER)
       return PAM_USER_UNKNOWN;
-    else if(pflags & ~PPW_SHADOW)
+    else if(pflags & ~(PPW_SHADOW|PPW_EXPIRING|PPW_WILLEXPIRE))
       return PAM_AUTHTOK_ERR;
     else
       return PAM_SUCCESS;
@@ -320,7 +320,7 @@ int pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv)
 
   for(i=0; i < argc; i++) {
      {
-	 char *tmp = xstrdup(argv[i]);
+	 char *tmp = x_strdup(argv[i]);
 	 strncpy(argbuf,strtok(tmp ,"="),255);
 	 strncpy(argval,strtok(NULL,"="),255);
 	 free(tmp);
@@ -398,7 +398,7 @@ int pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv)
 #endif
     pflags = _do_getpass(usrname,&curpass);
     if(curpass == NULL)
-      curpass = xstrdup("");
+      curpass = x_strdup("");
   }
   if(pflags & PPW_TOOEARLY) {
     conv_sendmsg(appconv,"You must wait longer to change your password",
@@ -569,7 +569,7 @@ STATIC int _do_setpass(char *forwho, char *towhat,int flags)
   setpwent();
   pwd = getpwnam(forwho);
 #ifdef DEBUG
-  printf("Got %X, for %s (salt %s)\n",pwd,
+  printf("Got %p, for %s (salt %s)\n",pwd,
 	 forwho,thesalt);
 #endif
   if(pwd == NULL)
@@ -724,7 +724,7 @@ STATIC int _do_getpass(char *forwho, char **theirpass)
     endspent();
     if(spwdent == NULL) 
       return PPW_NOSUCHUSER;
-    *theirpass = xstrdup(spwdent->sp_pwdp);
+    *theirpass = x_strdup(spwdent->sp_pwdp);
 
     /* We have the user's information, now let's check if their account
      has expired (60 * 60 * 24 = number of seconds in a day) */
@@ -744,19 +744,20 @@ STATIC int _do_getpass(char *forwho, char **theirpass)
 	&& (spwdent->sp_max != -1))
       /* Their passwd needs to be changed */
       retval |= PPW_EXPIRING;
-    else if((curdays > (spwdent->sp_lstchg + spwdent->sp_warn))
-	    && (spwdent->sp_warn != -1))
+    else if((curdays > (spwdent->sp_lstchg
+                        + spwdent->sp_max - spwdent->sp_warn))
+            && (spwdent->sp_max != -1) && (spwdent->sp_warn != -1))
       retval |= PPW_WILLEXPIRE;
 /*    if(spwdent->sp_lstchg < 0)
       retval &= ~(PPW_WILLEXPIRE | PPW_EXPIRING | PPW_EXPIRED);
     if(spwdent->sp_max < 0)
       retval &= ~(PPW_EXPIRING | PPW_EXPIRED); */
   } else {
-    *theirpass = (char *)xstrdup(pwd->pw_passwd);
+    *theirpass = (char *)x_strdup(pwd->pw_passwd);
   }
 
 #else
-  *theirpass = (char *) xstrdup(pwd->pw_passwd);
+  *theirpass = (char *) x_strdup(pwd->pw_passwd);
 #endif
 
   return retval;
@@ -766,6 +767,8 @@ STATIC int conv_sendmsg(struct pam_conv *aconv, const char *message, int style)
 {
   struct pam_message msg,*pmsg;
   struct pam_response *resp;
+  int retval;
+
   /* Get the password from the user... */
   pmsg          = &msg;
 
@@ -773,8 +776,12 @@ STATIC int conv_sendmsg(struct pam_conv *aconv, const char *message, int style)
   msg.msg = message;
   resp = NULL;
   
-  return aconv->conv(1, (const struct pam_message **) &pmsg,
-			 &resp, aconv->appdata_ptr);  
+  retval = aconv->conv(1, (const struct pam_message **) &pmsg,
+		       &resp, aconv->appdata_ptr);
+  if (resp) {
+      _pam_drop_reply(resp, 1);
+  }
+  return retval;
 }
 
 
@@ -785,6 +792,8 @@ STATIC int conv_getitem(struct pam_conv *aconv, char *message, int style,
   struct pam_response *resp;
   int retval;
 
+  D(("called."));
+
   /* Get the password from the user... */
   pmsg          = &msg;
   msg.msg_style = style;
@@ -802,17 +811,3 @@ STATIC int conv_getitem(struct pam_conv *aconv, char *message, int style,
   else
     return PAM_SERVICE_ERR;
 }
-
-/* Should add static support soon... */
-
-char *xstrdup(const char *x)
-{
-     char *s;
-
-     if ((s = (char *) malloc(strlen(x)+1))) {
-	  strcpy(s,x);
-     }
-     
-     return s;
-}
-
